EXTENSION_NAME = 'vertical-windows-list'

GObject = imports.gi.GObject
St = imports.gi.St
Gio = imports.gi.Gio
Meta = imports.gi.Meta
Clutter = imports.gi.Clutter
Shell = imports.gi.Shell
DND = imports.ui.dnd


ExtensionUtils = imports.misc.extensionUtils

Gettext = imports.gettext.domain(EXTENSION_NAME)
tr = Gettext.gettext

ExtensionUtils = imports.misc.extensionUtils
Main = imports.ui.main
PanelMenu = imports.ui.panelMenu
PopupMenu = imports.ui.popupMenu

class Extension
  constructor: ->
    @windowPanel = null


  enable: ->
    @monitor = Main.layoutManager.primaryMonitor

    @settings = ExtensionUtils.getSettings('org.gnome.shell.extensions.vertical-window-list')

    @windowPanel = new St.Widget
      style_class: 'windows-panel'
      reactive: true
      track_hover: true
      visible: true
      layout_manager: new Clutter.BinLayout()

    @settings.connect 'changed::panel-opacity', @opacityChanged.bind(@)
    @opacityChanged()

    @box = new St.BoxLayout
      x_expand: true
      y_expand: true
    @windowPanel.add_child @box
    @settings.connect 'changed::width', @widthChanged.bind(@)
    @widthChanged()

    @windowPanel.height = @monitor.height
    @windowPanel.set_position @monitor.x, 30

    @windowList = new St.Widget
      style_class: 'windows-list'
      reactive: true
      layout_manager: new Clutter.BoxLayout(homogeneous: true, orientation: Clutter.Orientation.VERTICAL)
      y_align: Clutter.ActorAlign.START
      x_expand: false
      y_expand: true

    @box.add_child @windowList
    @windowList.set_opacity_override(255)

    workspace = global.workspace_manager.get_active_workspace()

    for window in global.get_window_actors()
      @add_window(workspace, window.metaWindow)

    workspace.connect_after 'window-added', @add_window.bind(@)

    workspace.connect_after 'window-removed', @remove_window.bind(@)

    Main.layoutManager.addChrome @windowPanel, {affectsStruts: true, trackFullscreen: true}

    # cycle windows keybingings
    @cycling = false
    res = Main.wm.addKeybinding('cycle-up', @settings, Meta.KeyBindingFlags.NONE, Shell.ActionMode.NORMAL, @cycleWindow.bind(@, true))
    if res is Meta.KeyBindingAction.NONE
      log 'cycle-up keybingind failed'
    res = Main.wm.addKeybinding('cycle-down', @settings, Meta.KeyBindingFlags.NONE, Shell.ActionMode.NORMAL, @cycleWindow.bind(@, false))
    if res is Meta.KeyBindingAction.NONE
      log 'cycle-down keybingind failed'

  cycleWindow: (up, display, window, binding) ->
    # cycle up or down the list starting from previous Window
    buttons = @windowList.get_children()
    [activeWindow, ix] = ([b.metaWindow, ix] for b, ix in buttons when b.has_style_class_name('focused'))[0] or []
    if not activeWindow?
      return
    if @cycling is false
      @activateWindow(@previousWindow(activeWindow))
      @cycling = true
      return

    if up
      ix = if (ix == buttons.length-1) then 0 else ix + 1
    else
      ix = if (ix == 0) then buttons.length-1 else ix - 1

    @activateWindow(buttons[ix].metaWindow)
    @cycling = true



  widthChanged: ->
    @windowPanel.width = @settings.get_int('width')
  opacityChanged: ->
    @windowPanel.set_opacity @settings.get_int('panel-opacity')


  add_window: (workspace, metaWindow) ->
    for b in @windowList.get_children()
      if b.metaWindow is metaWindow
        return
      if metaWindow.is_skip_taskbar()
        return

    button = new St.Button
      style_class: 'window-button'
      can_focus: true
      x_expand: true
      button_mask: St.ButtonMask.ONE | St.ButtonMask.TWO | St.ButtonMask.THREE
    button.metaWindow = metaWindow

    button.connect 'clicked', @buttonClicked.bind(@, metaWindow)

    layout = new St.BoxLayout
      style_class: 'window-button-layout'
      x_expand: false
      y_expand: true
    button.set_child layout

    icon = new St.Bin(style_class: 'window-button-icon')
    layout.add icon
    texture_cache = St.TextureCache.get_default()
    texture_cache.connect 'icon-theme-changed', @updateIcon.bind(@, metaWindow, icon)

    metaWindow.connect_after 'notify::wm-class', @updateIcon.bind(@, metaWindow, icon)

    metaWindow.connect_after 'notify::gtk-application-id', @updateIcon.bind(@, metaWindow, icon)

    @settings.connect 'changed::icon-size', @updateIcon.bind(@, metaWindow, icon)
    @updateIcon metaWindow, icon

    button.title = new St.Label
      y_align: Clutter.ActorAlign.CENTER
      text: ''
      style_class: 'window-button-text'
    button.title.clutter_text.single_line_mode = true
    layout.add button.title
    signal1 = metaWindow.connect 'notify::title', @updateTitle.bind(@, metaWindow, button)
    button.connect 'destroy', -> metaWindow.disconnect signal1
    signal2 = global.display.connect 'notify::focus-window', @focusWindow.bind(@, button)
    button.connect 'destroy', -> global.display.disconnect signal2

    button.custom_title = new St.Entry
      y_align: Clutter.ActorAlign.CENTER
      text: ''
      style_class: 'window-button-editor'
      visible: false

    @updateTitle metaWindow, button

    layout.add button.custom_title

    button.custom_title.connect 'key-release-event', (custom_title, event) =>
      if event.get_key_code() in [36, 104] # Enter
        button.custom_title.visible = false
        button.title.visible = true
        @updateTitle(metaWindow, button)

    @windowList.add_child(button)

    # windows reordering via drag and drop
    button.draggable = DND.makeDraggable(button, {dragActorOpacity: 127})
    button.draggable.connect 'drag-end', @dragEnd.bind(@, button)
    button.draggable.connect 'drag-begin', ->
      dragActor = new St.Button
        style_class: 'drag-button'
        label: ''+metaWindow.title
      [sizeW, sizeH] = button.get_size()
      dragActor.set_size sizeW, sizeH
      [posX, posY] = button.get_transformed_position()
      dragActor.set_position  posX, posY
      dragActor.getDragActor = -> dragActor
      button._delegate = dragActor

  dragEnd: (button) ->
    dropRect = new Meta.Rectangle
      x: button.draggable._dragX
      y: button.draggable._dragY
      width:1
      height:1

    for b, ix in @windowList.get_children()
      bRect = new Meta.Rectangle
        x: b.get_transformed_position()[0]
        y: b.get_transformed_position()[1]
        width: b.get_transformed_size()[0]
        height: b.get_transformed_size()[1]

      if b is button
        continue

      if bRect.contains_rect(dropRect)
        @windowList.remove_child(button)
        @windowList.insert_child_at_index(button, ix)

    # drag under last window
    if dropRect.y > bRect.y
      @windowList.remove_child(button)
      @windowList.add_child(button)


  remove_window: (workspace, metaWindow) ->
    for button in @windowList.get_children()
      if button.metaWindow is metaWindow
        button.destroy()


  focusWindow: (button) ->
    if global.display.focus_window is button.metaWindow
      button.add_style_class_name('focused');
    else
      button.remove_style_class_name('focused');


  updateTitle: (metaWindow, button) ->
    button.title.set_text button.custom_title.text or ('' + metaWindow.title)


  app: (metaWindow) -> Shell.WindowTracker.get_default().get_window_app(metaWindow)


  updateIcon: (metaWindow, icon) ->
    icon.child = @app(metaWindow)?.create_icon_texture(@settings.get_int('icon-size')) or new St.Icon({ icon_name: 'icon-missing', icon_size: @settings.get_int('icon-size') })


  buttonClicked: (metaWindow, button, button_number) ->
    # activate previous window if already was active
    if button_number is 1
      if metaWindow.appears_focused
        prevWindow = @previousWindow(metaWindow)
        @activateWindow(prevWindow)
      else
        @activateWindow(metaWindow)
      return

    if button_number is 2
      button.title.visible = false
      button.custom_title.visible = true
      button.custom_title.x_expand = true
      button.custom_title.grab_key_focus()
      return

    # close window on right click
    if button_number is 3
      @activateWindow(metaWindow)
      metaWindow.delete(global.get_current_time())
      return


  activateWindow: (metaWindow) ->
    metaWindow.activate(global.get_current_time())
    @cycling = false

  previousWindow: (metaWindow) ->
    return global.display.get_tab_next(Meta.TabList.NORMAL, metaWindow.get_workspace(), metaWindow, false)

  disable: ->
    @windowPanel.destroy()
    @windowPanel = null

init = ->
  new Extension()

